@@ -3,8 +3,10 @@
 '''Object wrapper for key-value db'''
 import os
 import sys
-import marshal
 import time
+import copy
+import inspect
+import marshal
 from datetime import datetime, date, timedelta
 
 
@@ -34,15 +36,16 @@ class Model(object):
     ...     created = DateTimeProperty(key='c')
     ...     item    = ChoiceProperty(choices=[1, 2, 3])
     ...     item1   = ChoiceProperty(choices=(('a', 'val1'), ('b', 'val2')))
+    ...     hash    = Property(default={})
     ...
     ...     __db_type__ = 'dict'
     
     >>> Test(
     ...     day     = date(2000, 01, 02),
-    ...     foo     = [1, u'ы', {'a': 'b'}],
-    ...     ustr    = u'ыыы',
-    ...     cstr    = 'A' * 1000,
-    ...     cuni    = u'Ы' * 1000,
+    ...     foo     = [1, u'a', {'a': 'b'}],
+    ...     ustr    = u'bbb',
+    ...     cstr    = 'ccc' * 1000,
+    ...     cuni    = u'eee' * 1000,
     ...     created = datetime.now(),
     ... ).save()
     >>> Test(day=date(1999, 02, 02)).save()
@@ -53,9 +56,15 @@ class Model(object):
     >>> t.item = 2
     >>> t.item1 = 'val2'
     
+    >>> t.hash['a'] = 'b'
+    >>> t.save()
+    >>> Test.get(date(1999, 02, 02)).hash
+    {'a': 'b'}
+    
+    
     >>> Test.count()
     2
-    >>> Test.count(lambda x: x.ustr == u'ыыы')
+    >>> Test.count(lambda x: x.ustr == u'bbb')
     1
     
     >>> [t.day.year for t in Test.find(
@@ -71,6 +80,7 @@ class Model(object):
     __db_type__ = 'tc'
     
     def __init__(self, _raw_dict=None, **kwargs):
+        self._cached_data = {}
         if _raw_dict:
             self._data = _raw_dict
         else:
@@ -84,12 +94,15 @@ class Model(object):
         if pk is None:
             assert 0, 'not set primary key "%s" on model "%s"' % (
                 self._pk_name, self.__class__.__name__)
+        for prop, value in self._cached_data.iteritems():
+            if isinstance(value, (dict, list)):
+                self._data[prop.key] = prop.encode(value)
         self.db[md(pk)] = md(self._data)
         self._data['__pk__'] = pk
         
     @classmethod
     def get(cls, key):
-        key = cls.__dict__[cls._pk_name].encode(key)
+        key = cls._properties[cls._pk_name].encode(key)
         try:
             kw = cls.db[md(key)]
         except KeyError:
@@ -176,7 +189,7 @@ class Model(object):
     
     @cached_classmethod
     def _pk_name(cls):
-        ret = [k for k, v in cls._properties.iteritems() if v == '__pk__']
+        ret = [k for k, v in cls._properties.iteritems() if v.key == '__pk__']
         if not ret:
             assert 0, 'primary key in model "%s" not found' % cls.__name__
         return ret[0]
@@ -184,28 +197,26 @@ class Model(object):
     @cached_classmethod
     def _properties(cls):
         '''
-        {prop_attr_name: save_key_name}
+        {prop_name: prop}
         '''
-        def req_bases(cls):
-            bases = list(cls.__bases__) or []
-            ret = []
-            for base in bases:
-                for c in req_bases(base):
-                    if c not in ret and c not in bases:
-                        ret.append(c)
-            ret.extend(bases)
-            return ret        
         ret = {}
-        for k, prop in cls.__dict__.iteritems():
-            if BaseProperty in req_bases(prop.__class__):
+        all_members = {}
+        for c in inspect.getmro(cls):
+            for k, v in c.__dict__.iteritems():
+                if k not in all_members:
+                    all_members[k] = v
+        keys = []
+        for k, prop in all_members.iteritems():
+            if issubclass(prop.__class__, BaseProperty):
                 if not prop.key:
                     prop.key = k
-                if prop.key in ret.values():
+                if prop.key in keys:
                     keyname = 'primary key' if prop.key == '__pk__' else 'key "%s"' % prop.key
                     assert 0, 'multiply %s in model "%s"' % (keyname, cls.__name__)
-                ret[k] = prop.key
+                ret[k] = prop
+                keys.append(prop.key)
         return ret
-        
+
     def __unicode__(self):
         keys = self._properties.keys()
         values = {}
@@ -240,15 +251,26 @@ class BaseProperty(object):
         self.primary_key = primary_key
         
     def __get__(self, instance, owner):
+        if self in instance._cached_data:
+            return instance._cached_data[self]
         if self.key in instance._data:
-            return self.decode(instance._data[self.key])
-        return self.default
-        
+            ret = self.decode(instance._data[self.key])
+        elif isinstance(self.default, (dict, list)):
+            ret = copy.copy(self.default)
+        else:
+            ret = self.default
+        instance._cached_data[self] = ret
+        return ret
+
     def __set__(self, instance, value):
+        instance._cached_data[self] = value
         instance._data[self.key] = self.encode(value)
         
     def __delete__(self, instance):
-        del instance._data[self.key]
+        if self in instance._cached_data:
+            del instance._cached_data[self]
+        if self.key in instance._data:
+            del instance._data[self.key]
     
     def encode(self, value):
         '''Приведение к маршализируемому типу'''
